// $Id: CoulombAction.cc,v 1.41 2008/03/01 19:48:58 jshumwa Exp $
/*  Copyright (C) 2004-2006 John B. Shumway, Jr.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include "CoulombAction.h"
#include "MultiLevelSampler.h"
#include "Beads.h"
#include "Paths.h"
#include "SuperCell.h"
#include "SimulationInfo.h"
#include "PairAction.h"
#include "ImagePairAction.h"
#include "EwaldSum.h"

CoulombAction::CoulombAction(const double epsilon, 
    const SimulationInfo& simInfo, const int norder, double rmin, 
    double rmax, int ngpts, const bool dumpFiles,
    bool useEwald, int ewaldNDim, double ewaldKcut, double screenDist)
  : epsilon(epsilon), tau(simInfo.getTau()), npart(simInfo.getNPart()),
    pairActionArray(0), screenDist(screenDist), ewaldSum(0) {
  typedef blitz::TinyVector<int, NDIM> IVec;
  const int nspecies=simInfo.getNSpecies();
  if (useEwald && NDIM==3 && ewaldNDim==NDIM) {
    SuperCell &cell(*simInfo.getSuperCell());
    double rcut = cell.a[0]/2.;
    ewaldSum = new EwaldSum(cell,npart,rcut,ewaldKcut);
    rewald.resize(npart);
    EwaldSum::Array1 &q=ewaldSum->getQArray();  
    for (int i=0; i<npart; ++i) q(i)=simInfo.getPartSpecies(i).charge;
    ewaldSum->evalSelfEnergy();
    std::cout << "using ewald sums" << ewaldSum << std::endl;
  }
  for (int i=0; i<nspecies; ++i) {
    for (int j=i; j<nspecies; ++j) {
      const Species &s1=simInfo.getSpecies(i);
      const Species &s2=simInfo.getSpecies(j);
      q1q2=s1.charge*s2.charge/epsilon;
      if (q1q2!=0 && ((s1.name!=s2.name) || s1.count>1) ) {
        mu=1./(1./s1.mass+1./s2.mass);
        /// Hack to handle ion-ion interaction properly.
        if (rmin==0) rmin=0.005/(2*((mu>500)?1:mu)*fabs(q1q2)); 
        IVec nimage=0; bool needImages=false;
        if (useEwald && ewaldNDim<NDIM) {
          SuperCell::Vec a=simInfo.getSuperCell()->a;
          for (int idim=0; idim<ewaldNDim; ++idim) {
            nimage[idim]=(int)(rmax/a[idim]);
            if (nimage[idim]>1) nimage[idim]-=2;
            if (nimage[idim]>0) needImages=true;
          }
        } else {
          if (rmax==0) {
            SuperCell::Vec a=simInfo.getSuperCell()->a;
            rmax=sqrt(dot(a,a));
          }
        }
        if (ngpts==0) ngpts=500;
        if (needImages) {
          pairActionArray.push_back(
            /// Hack to handle ion-ion interaction properly.
            new ImagePairAction(s1,s2, *this, simInfo, (mu>500)?0:norder,
                           nimage, rmin, rmax, ngpts));
        } else {
          pairActionArray.push_back(
            /// Hack to handle ion-ion interaction properly.
            new PairAction(s1,s2, *this, simInfo, (mu>500)?0:norder,
                           rmin, rmax, ngpts));
        }
        if (dumpFiles) (*(pairActionArray.end()-1))->write("");
      }
    }
  }
  std::cout << "Coulomb action" << std::endl;
}

CoulombAction::~CoulombAction() {
  for (unsigned int i=0; i<pairActionArray.size(); ++i) {
     delete pairActionArray[i];
  }
  delete ewaldSum;
}

double CoulombAction::getActionDifference(const MultiLevelSampler& sampler,
                                         const int level) {
  double u=0;
  for (unsigned int i=0; i<pairActionArray.size(); ++i) {
    u += pairActionArray[i]->getActionDifference(sampler,level);
  }
  // Compute long range ewald action at lowest level.
  if (ewaldSum && level==0) {
    const Beads<NDIM>& sectionBeads=sampler.getSectionBeads();
    const Beads<NDIM>& movingBeads=sampler.getMovingBeads();
    const int nSlice=sectionBeads.getNSlice();
    const IArray& index=sampler.getMovingIndex(); 
    const int nMoving=index.size();
//std::cout << "    u=" << u << std::endl;
    for (int islice=1; islice<nSlice-1; ++islice) {
      for (int i=0; i<npart; ++i) rewald(i)=sectionBeads(i,islice);
      u -= ewaldSum->evalLongRange(rewald)*tau/epsilon;
//      std::cout << ewaldSum->evalLongRange(rewald)*tau/epsilon << std::endl;
      for (int i=0; i<nMoving; ++i) rewald(index(i))=movingBeads(i,islice);
      u += ewaldSum->evalLongRange(rewald)*tau/epsilon;
//      std::cout << ewaldSum->evalLongRange(rewald)*tau/epsilon << std::endl;
    }
//std::cout << "now u=" << u << std::endl;
  }
  return u;
}

double CoulombAction::getTotalAction(const Paths& paths, int level) const {
  return 0;
}

void CoulombAction::getBeadAction(const Paths& paths, int ipart, int islice,
         double& u, double& utau, double& ulambda, Vec& fm, Vec& fp) const {
  u=utau=ulambda=0; fm=0.; fp=0.;
  for (unsigned int i=0; i<pairActionArray.size(); ++i) {
    double ui=0, utaui=0, ulambdai=0; 
    Vec fmi=0.0, fpi=0.0; 
    pairActionArray[i]->
      getBeadAction(paths,ipart,islice,ui,utaui,ulambdai,fmi,fpi);
    u+=ui; utau+=utaui; ulambda+=ulambdai; fm+=fmi; fp+=fpi;
  }
  // Compute long range action.
  if (ewaldSum && ipart==0) {
    paths.getSlice(islice,rewald);
    double longRange = ewaldSum->evalLongRange(rewald)/epsilon;
    u += longRange*tau;
    utau += longRange;
  }

}

double CoulombAction::getEField(const Paths& paths, int ipart, 
                                int islice) const {
  double u=0,utau=0,ulambda=0;
  Vec fm=0.,fp=0.;
  for (unsigned int i=0; i<pairActionArray.size(); ++i) {
    double ui=0, utaui=0, ulambdai=0; 
    Vec fmi=0.0, fpi=0.0; 
    pairActionArray[i]->
      getBeadAction(paths,ipart,islice,ui,utaui,ulambdai,fmi,fpi);
    u+=ui; utau+=utaui; ulambda+=ulambdai; fm+=fmi; fp+=fpi;
  }
  return fp[0]/tau;
}

double CoulombAction::u(double r, int order) const {
  double taueff = 2.0*mu*q1q2*q1q2*tau;
  double u=0;
#if NDIM==3
  double reff = 2.0*mu*q1q2*r;
  double stau = q1q2*sqrt(2.0*mu*tau);
  if (order==0) {
    //u = u0 - log(a);
    double u0 = stau*(1.772453851+stau*(-0.074137740+stau*(0.005834805
               +stau*(-0.000382686+stau*(0.000008738+stau*0.000002138)))));
    double a = 0.81644994/stau + 0.20750388;
    double b = 0.73884476/taueff + 0.22874030/stau;
    double c = 0.48135991*pow(stau,-3) + 0.15828689*pow(stau,-2);
    double d = 0.15769661*pow(stau,-4) + 0.01512610*pow(stau,-3);
    double e = 0.25472475*pow(stau,-5) + 0.16019595*pow(stau,-4);
    double f = 0.51100692/stau + 0.15237602;
    double g = 0.15705841/taueff + 0.01503638/stau;
    u = (u0 + reff*( (-1.+u0*a) + reff*(f + reff*(g + e*taueff*reff)))) /
        (1. + reff*(a + reff*(b + reff*(c + reff*(d + e*reff)))));
    if (ewaldSum) {
      u += tau*q1q2*ewaldSum->evalShortRange(r);
    }
  } else if (order==1) {
    //u = u1;// - b/a;
    double a =  0.23707617/stau - 0.05259752;
    double b =  0.02975747/taueff - 0.04859804/stau;
    double c =  4.73092072/stau - 3.11503429;
    double d = -3.83954107/taueff + 3.31549995/stau;
    double e =  3.64988162*pow(stau,-3) - 1.54310812/taueff;
    double f =  0.31095112*pow(stau,-4) - 0.49311207*pow(stau,-3);
    u=reff*reff*(a+b*reff)/(1+reff*(c+reff*(d+reff*(e+reff*f))));
  } else if (order==2) {
    //u = u2; - (a*c-b*b)/(a*a);
    //double a =  0.00359686872*pow(stau,-2) + 0.000041584197*pow(stau,-1);
    //double b =  0.272325152*pow(stau,-4) + 0.0272612295*pow(stau,-3);
    //u=a*pow(reff,3)/(1+b*pow(reff,4)); 
    double a = 0.00486183169*pow(stau,-3)+0.0031962421*pow(stau,-2);
    double b =-2.44343134*pow(stau,-1)+0.011967222;
    double c = 5.34200782*pow(stau,-2)+0.413911437*pow(stau,-1);
    double d =-3.09120046*pow(stau,-3)+0.592511039*pow(stau,-2);
    double e = 0.625141209*pow(stau,-4)-0.328624631*pow(stau,-3);
    double f = 0.350171848*pow(stau,-5)+0.286023259*pow(stau,-4);
    u=a*pow(reff,4)/(1+reff*(b+reff*(c+reff*(d+reff*(e+reff*f)))));
  } else if (order==3) {
    u = 0;//u3; - (6*a*a*d-6*a*b*c+b*b*b)/(a*a*a);
  }
#endif
#if NDIM==2
  double ureff = 2.0*mu*fabs(q1q2)*r;
  double sign= (q1q2>0)?1:-1;
  if (order==0) {
    double u0=0;
    const double p[]={0.,2.784163998,-0.331414576,0.060113708,-0.007701621};
    for (int i=1; i<5; ++i) {
      u0+=pow(sign,i)*p[i]*pow(taueff,0.5*i);
    }
    double up0=-2*sign;
    double w1=taueff*sign;
    double w3=sign*pow(taueff,2)/6.;
    //double w4=-pow(taueff,3)/12.0;
    double a=u0;
    double f=1;
    //double f=(w1*w1*w1-w3*u0-w1*w3*up0)
    //        /(w1*w1*w4-w3*w3*u0);
    double c=w1*f;
    double d=(a-f*w3)/w1;
    double b=up0+a*d;
    double e=f*b/c;
    u=(a+b*ureff+c*ureff*ureff)/(1+d*ureff+e*ureff*ureff+f*ureff*ureff*ureff);
    if (screenDist!=0) {
      ureff=2.0*mu*fabs(q1q2)*sqrt(r*r+screenDist*screenDist);
      u-=(a+b*ureff+c*ureff*ureff)/(1+d*ureff+e*ureff*ureff+f*ureff*ureff*ureff);
    }
  }
#endif
  return u;
}

double CoulombAction::utau(double r, int order) const {
  double tausave=tau;
  const double eps=1e-7;
  tau = tausave*(1.0+eps);
  double up=u(r,order);
  tau = tausave*(1.0-eps);
  double um=u(r,order);
  return (up-um)/(2.0*eps*tau);
}
