// $Id: GridPotential.cc,v 1.11 2006/10/18 17:08:18 jshumwa Exp $
/*  Copyright (C) 2004-2006 John B. Shumway, Jr.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include "GridPotential.h"
#include <blitz/tinyvec-et.h>
#include "MultiLevelSampler.h"
#include "SimulationInfo.h"
#include "Beads.h"
#include "Paths.h"
#include "SuperCell.h"
#include <H5Cpp.h>
#include <fstream>

GridPotential::GridPotential(const SimulationInfo& simInfo,
                             const std::string& filename)
  : tau(simInfo.getTau()), 
    vindex(simInfo.getNPart(),(Array3*)0){
  H5::H5Library::open();
  H5::Exception::dontPrint();
  int nn;
  vgrid.resize(2);
  try {
    // Read the bandoffsets from grid.h5.
    H5::H5File file(filename,H5F_ACC_RDONLY);
    H5::Group group = file.openGroup(filename=="emagrids.h5"?"boffset":"/"); {
      H5::DataSet dataset = group.openDataSet("vh");
      hsize_t dims[3];
      dataset.getSpace().getSimpleExtentDims(dims);
      n1=dims[0]; n2=dims[1]; n3=dims[2]; nn=n1*n2*n3;
      vgrid[0].resize(n1,n2,n3);
      dataset.read(vgrid[0].data(),H5::PredType::NATIVE_DOUBLE);
      vgrid[1].resize(n1,n2,n3);
      group.openDataSet("ve").read(vgrid[1].data(),H5::PredType::NATIVE_DOUBLE);
    } { // Read the grid spacing.
      double a=0;
      group.openDataSet("a").read(&a,H5::PredType::NATIVE_DOUBLE);
      b=1.0/a;
    } 
  } catch (H5::Exception& error) { 
    error.printError();
  }
  H5::H5Library::close();
  // Convert grids from eV to Ha.
  const double eVtoHa=1./27.211396;
  if (filename=="boffset.h5") {
    for (unsigned int i=0; i<vgrid.size(); ++i) {vgrid[i]*=eVtoHa;}
  } 
  // Flip sign of hole grid if read from emagrid.h5
  if (filename=="emagrids.h5") vgrid[0]*=-1; 
  // Setup the vindex.
  for (int i=0; i<simInfo.getNPart(); ++i) {
    std::string name=simInfo.getPartSpecies(i).name;
    if (name.substr(0,1)=="h") vindex[i]=&vgrid[0];
    else if (name.substr(0,1)=="e") vindex[i]=&vgrid[1];
  }
}

double GridPotential::getActionDifference(const MultiLevelSampler& sampler,
                                         const int level) {
  const Beads<NDIM>& sectionBeads=sampler.getSectionBeads();
  const Beads<NDIM>& movingBeads=sampler.getMovingBeads();
  const SuperCell& cell=sampler.getSuperCell();
  const int nStride=(int)pow(2,level);
  const int nSlice=sectionBeads.getNSlice();
  const IArray& index=sampler.getMovingIndex(); 
  const int nMoving=index.size();
  double deltaAction=0;
  // Read values off of the grid.
  //for (int islice=nStride; islice<nSlice-nStride; islice+=nStride) {
  for (int islice=0; islice<nSlice-nStride; islice+=nStride) {
    for (int iMoving=0; iMoving<nMoving; ++iMoving) {
      const int i=index(iMoving);
      // Add action for moving beads. (Evaluate v at midpoint)
      Vec r=movingBeads(iMoving,islice);
      cell.pbc(r);
      Vec delta=movingBeads(iMoving,islice+nStride);
      delta-=r; cell.pbc(delta);delta*=0.5; r+=delta;
      cell.pbc(r);
      deltaAction+=v(r,i)*tau*nStride;
      // Subtract action for old beads.
      r=sectionBeads(i,islice);
      cell.pbc(r);
      delta=sectionBeads(i,islice+nStride);
      delta-=r; cell.pbc(delta);delta*=0.5; r+=delta;
      cell.pbc(r);
      deltaAction-=v(r,i)*tau*nStride;
    }
  }
  return deltaAction;
}

double GridPotential::v(Vec r, const int i) const {
  r*=b;
  int i1=(int)floor(r[0]), i2=(int)floor(r[1]), i3=(int)floor(r[2]);
  double x=r[0]-i1, y=r[1]-i2, z=r[2]-i3;
  i1+=n1/2; i2+=n2/2; i3+=n3/2;
  if (i1<0) {i1=0; x=0;}; if (i2<0) {i2=0; y=0;}; if (i3<0) {i3=0; z=0;};
  if (i1>n1-2) {i1=n1-2; x=1;};
  if (i2>n2-2) {i2=n2-2; y=1;};
  if (i3>n3-2) {i3=n3-2; z=1;};
  const Array3& v(*vindex[i]);
  double V =(1-z)*( (1-y)*( (1-x)*v(i1,i2,i3)+x*v(i1+1,i2,i3) ) 
                     + y *( (1-x)*v(i1,i2+1,i3)+x*v(i1+1,i2+1,i3) ) )
             + z *( (1-y)*( (1-x)*v(i1,i2,i3+1)+x*v(i1+1,i2,i3+1) ) 
                     + y *( (1-x)*v(i1,i2+1,i3+1)+x*v(i1+1,i2+1,i3+1)));
  return V;
}

double GridPotential::getTotalAction(const Paths& paths, int level) const {
  return 0;
}


void GridPotential::getBeadAction(const Paths& paths, int ipart, int islice,
    double& u, double& utau, double& ulambda, Vec& fm, Vec& fp) const {
  const Array3& v(*vindex[ipart]);
  Vec r=paths(ipart,islice);
  //int jslice=islice-1; if (jslice<0) jslice=beads.getNSlice()-1;
  //r+=beads(ipart,jslice); r*=0.5;
  r*=b;
  int i1=(int)floor(r[0]), i2=(int)floor(r[1]), i3=(int)floor(r[2]);
  double x=r[0]-i1, y=r[1]-i2, z=r[2]-i3;
  i1+=n1/2; i2+=n2/2; i3+=n3/2; 
  if (i1<0) {i1=0; x=0;} if (i2<0) {i2=0; y=0;} if (i3<0) {i3=0; z=0;}
  if (i1>n1-2) {i1=n1-2; x=1;}
  if (i2>n2-2) {i2=n2-2; y=1;}
  if (i3>n3-2) {i3=n3-2; z=1;}
  utau =(1-z)*( (1-y)*( (1-x)*v(i1,i2,i3)+x*v(i1+1,i2,i3) ) 
                 + y *( (1-x)*v(i1,i2+1,i3)+x*v(i1+1,i2+1,i3) ) )
         + z *( (1-y)*( (1-x)*v(i1,i2,i3+1)+x*v(i1+1,i2,i3+1) ) 
                 + y *( (1-x)*v(i1,i2+1,i3+1)+x*v(i1+1,i2+1,i3+1)));
  u = utau*tau;
  ulambda=0;
  fm[0] =(1-z)*( (1-y)*( (1-x)*v(i1,i2,i3)+x*v(i1+1,i2,i3) )
                  + y *( (1-x)*v(i1,i2+1,i3)+x*v(i1+1,i2+1,i3) ) )
          + z *( (1-y)*( (1-x)*v(i1,i2,i3+1)+x*v(i1+1,i2,i3+1) ) 
                  + y *( (1-x)*v(i1,i2+1,i3+1)+x*v(i1+1,i2+1,i3+1)));
  fm[1] =(1-z)*( (1-y)*( (1-x)*v(i1,i2,i3)+x*v(i1+1,i2,i3) ) 
                  + y *( (1-x)*v(i1,i2+1,i3)+x*v(i1+1,i2+1,i3) ) )
          + z *( (1-y)*( (1-x)*v(i1,i2,i3+1)+x*v(i1+1,i2,i3+1) ) 
                  + y *( (1-x)*v(i1,i2+1,i3+1)+x*v(i1+1,i2+1,i3+1)));
  fm[2] =(1-z)*( (1-y)*( (1-x)*v(i1,i2,i3)+x*v(i1+1,i2,i3) ) 
                  + y *( (1-x)*v(i1,i2+1,i3)+x*v(i1+1,i2+1,i3) ) )
          + z *( (1-y)*( (1-x)*v(i1,i2,i3+1)+x*v(i1+1,i2,i3+1) ) 
                  + y *( (1-x)*v(i1,i2+1,i3+1)+x*v(i1+1,i2+1,i3+1)));
  fm*=0.5;
  fp=fm;
}
