// $Id: PairGridAction.cc,v 1.3 2006/10/18 17:08:19 jshumwa Exp $
/*  Copyright (C) 2004-2006 John B. Shumway, Jr.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include "PairGridAction.h"
#include "MultiLevelSampler.h"
#include "Beads.h"
#include "Paths.h"
#include "SuperCell.h"
#include "SimulationInfo.h"
#include <vector>
#include <fstream>
#include <H5Cpp.h>
#include <blitz/tinyvec-et.h>

PairGridAction::PairGridAction(const Species& s1, const Species& s2,
            const std::string& filename, const SimulationInfo& simInfo) 
  : tau(simInfo.getTau()), ifirst1(s1.ifirst), ifirst2(s2.ifirst),
    npart1(s1.count), npart2(s2.count){ 
 std::cout << "constructing PairGridAction" << std::endl;
 std::cout << "species1= " <<  s1 << "species2= " <<  s2 << std::endl;
 std::cout << "filename=" << filename << std::endl;
 H5::H5Library::open();
 H5::Exception::dontPrint();
 int nn;
 ugrid.resize(2);
 try{
   H5::H5File file(filename, H5F_ACC_RDONLY);
   H5::Group group=file.openGroup("Fits_6"); {
     H5::DataSet dataset = group.openDataSet("Fit_9/Umat");
     hsize_t dims[2];
     dataset.getSpace().getSimpleExtentDims(dims);
     n1=dims[0]; n2=dims[1]; nn=n1*n2;
     ugrid[0].resize(n1,n2);
     dataset.read(ugrid[0].data(), H5::PredType::NATIVE_DOUBLE);
     ugrid[1].resize(n1,n2);
     group.openDataSet("Fit_9/dUmat").read(ugrid[1].data(), H5::PredType::NATIVE_DOUBLE);
   } { double delq=0, qE=0, qNP=0, qS=0;
       double delt=0, tE=0, tNP=0, tS=0;
       group.openDataSet("qGrid_0/End").read(&qE,H5::PredType::NATIVE_DOUBLE); 
       group.openDataSet("qGrid_0/NumPoints").read(&qNP,H5::PredType::NATIVE_DOUBLE);
       group.openDataSet("qGrid_0/Start").read(&qS,H5::PredType::NATIVE_DOUBLE);  
       delq=(qE-qS)/(qNP-1.0);
       delqinv=1.0/delq;
       group.openDataSet("tGrid_1/End").read(&tE,H5::PredType::NATIVE_DOUBLE); 
       group.openDataSet("tGrid_1/NumPoints").read(&tNP,H5::PredType::NATIVE_DOUBLE);
       group.openDataSet("tGrid_1/Start").read(&tS,H5::PredType::NATIVE_DOUBLE);  
       delt=(tE-tS)/(tNP-1.0);
       deltinv=1.0/delt;
     }
} 
 catch (H5::Exception& error){
   error.printError();
 }
 H5::H5Library::close();
}

double PairGridAction::getActionDifference(const MultiLevelSampler& sampler,
                                         const int level) {
  const Beads<NDIM>& sectionBeads=sampler.getSectionBeads();
  const Beads<NDIM>& movingBeads=sampler.getMovingBeads();
  const SuperCell& cell=sampler.getSuperCell();
  const int nStride=(int)pow(2,level);
  const int nSlice=sectionBeads.getNSlice();
  const IArray& index=sampler.getMovingIndex(); 
  const int nMoving=index.size();
  double deltaAction=0;
  for (int iMoving=0; iMoving<nMoving; ++iMoving) {
    const int i=index(iMoving);
    int jbegin,jend;
    if (i>=ifirst1 && i<ifirst1+npart1) {
      jbegin=ifirst2; jend=ifirst2+npart2;
    } else
    if (i>=ifirst2 && i<ifirst2+npart2) {
      jbegin=ifirst1; jend=ifirst1+npart1;
    } else 
    continue; //Particle not in this interaction.
    for (int j=jbegin; j<jend; ++j) {
      bool isMoving=false; int jMoving=0;
      for (int k=0;k<nMoving;++k) {
        if (j==index(k)) {isMoving=true; jMoving=k; break;}
      }
      if (isMoving && i<=j) continue; //Don't double count moving interactions.
      Vec prevDelta =sectionBeads(i,0);
      prevDelta-=sectionBeads(j,0); cell.pbc(prevDelta);
      Vec prevMovingDelta=prevDelta;
      double prevR=sqrt(dot(prevDelta,prevDelta));
      double prevMovingR=prevR;
      for (int islice=nStride; islice<nSlice; islice+=nStride) {
        // Add action for moving beads.
        Vec delta=movingBeads(iMoving,islice);
        delta-=(isMoving)?movingBeads(jMoving,islice):sectionBeads(j,islice);
        cell.pbc(delta);
        double r=sqrt(dot(delta,delta));
        double q=0.5*(r+prevMovingR);
        Vec svec=delta-prevMovingDelta;
	double magsvec=sqrt(dot(svec,svec));
       	double t=0.5*magsvec/q;
 // std::cout<<"PG1: r=" << r << " q=" << q   ;
        deltaAction += v(q,t);
 // std::cout << " PG delAct1 = " << deltaAction << std::endl;
        prevMovingDelta=delta; prevMovingR=r;
        
	// Subtract action for old beads.
        delta=sectionBeads(i,islice);
        delta-=sectionBeads(j,islice);
        cell.pbc(delta);
        r=sqrt(dot(delta,delta));
        q=0.5*(r+prevR);
        svec=delta-prevDelta;
       	magsvec=sqrt(dot(svec,svec));
	t=0.5*magsvec/q;
 // std::cout<<"PG2: r=" << r << " q=" << q ;
	deltaAction-= v(q,t);
 // std::cout << " PG delAct2 = " << deltaAction << std::endl;
        prevDelta=delta; prevR=r;
      }
    }
  }
  return deltaAction*=nStride;
} 

double PairGridAction::getTotalAction(const Paths& paths, int level) const {
  return 0;
}

void PairGridAction::getBeadAction(const Paths& paths, int ipart, int islice,
         double& u, double& utau, double& ulambda, Vec& fm, Vec& fp) const {
  u=utau=ulambda=0; fm=0.; fp=0.;
  int jbegin,jend;
  if (ipart>=ifirst1 && ipart<ifirst1+npart1) {
    jbegin=ifirst2; jend=ifirst2+npart2;
  } else
  if (ipart>=ifirst2 && ipart<ifirst2+npart2) {
    jbegin=ifirst1; jend=ifirst1+npart1;
  } else 
  return; //Particle not in this interaction.
  for (int j=jbegin; j<jend; ++j) {
    if (ipart==j) continue;
    Vec delta=paths(ipart,islice);
    delta-=paths(j,islice);
    paths.getSuperCell().pbc(delta);
    double r=sqrt(dot(delta,delta));
    Vec prevDelta=paths(ipart,islice,-1);
    prevDelta-=paths(j,islice,-1);
    paths.getSuperCell().pbc(prevDelta);
    double rp=sqrt(dot(prevDelta,prevDelta));
    double q=0.5*(r+rp);
    Vec svec=delta-prevDelta; double t=0.5*sqrt(dot(svec,svec))/q;
    u += 0.5*v(q,t);
    utau += 0.5*dv(q,t);
// std::cout << "PG q=" << q << " u= "<< u << " utau " << utau << std::endl;
  }
}

double PairGridAction::v(double q, double t) const {
  double V=0;
  q*=delqinv, t*=deltinv;
  int i=(int)floor(q), j=(int)floor(t);
  double x=q-i, y=t-j;
  if (i<0) {i=0; x=0;}; if (j<0) {j=0; y=0;} ;
  if (i>n1-2) {i=n1-2; q=1;};
  if (j>n2-2) {j=n2-2; t=1;};
   V= (1-y)*((1-x)*ugrid[0](i,j)+x*ugrid[0](i+1,j)) 
        + y*((1-x)*ugrid[0](i,j+1)+x*ugrid[0](i+1,j+1) ) ;
  return V;
}

double PairGridAction::dv(double q, double t) const {
  double dV=0;
  q*=delqinv, t*=deltinv;
  int i=(int)floor(q), j=(int)floor(t);
  double x=q-i, y=t-j;
  if (i<0) {i=0; x=0;}; if (j<0) {j=0; y=0;};  
  if (i>n1-2) {i=n1-2; q=1;};
  if (j>n2-2) {j=n2-2; t=1;};
  dV = (1-y)*( (1-x)*ugrid[1](i,j)+x*ugrid[1](i+1,j) ) 
        + y *( (1-x)*ugrid[1](i,j+1)+x*ugrid[1](i+1,j+1) ) ;
  return dV;
}
