//$Id: WireNodes.cc,v 1.9 2008/04/25 01:46:51 jshumwa Exp $
/*  Copyright (C) 2004-2008 John B. Shumway, Jr.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <blitz/tinyvec-et.h>
#include <tvmet/Matrix.h>
#include <tvmet/Vector.h>
#include "WireNodes.h"
#include "PeriodicGaussian.h"
#include "SimulationInfo.h"
#include "Species.h"
#include "SuperCell.h"
#include "Beads.h"
#include "DoubleMLSampler.h"

#define DGETRF_F77 F77_FUNC(dgetrf,DGETRF)
extern "C" void DGETRF_F77(const int*, const int*, double*, const int*,
                           const int*, int*);
#define DGETRI_F77 F77_FUNC(dgetri,DGETRI)
extern "C" void DGETRI_F77(const int*, double*, const int*, const int*,
                           double*, const int*, int*);

WireNodes::WireNodes(const SimulationInfo &simInfo, const Species &species,
    const double omega, const double temperature, const int maxlevel,
    const bool useUpdates, const int maxMovers)
  : NodeModel("_"+species.name),
    tau(simInfo.getTau()),mass(species.mass),npart(species.count),
    ifirst(species.ifirst), 
    matrix((int)(pow(2,maxlevel)+0.1)+1),
    ipiv(npart),lwork(npart*npart),work(lwork),
    cell(*simInfo.getSuperCell()), pg(0), pgp(0), pgm(0),
    omega(omega), coshwt(cosh(omega*0.5/temperature)),
    sinhwt(sinh(omega*0.5/temperature)), c(mass*0.5*omega/sinhwt),
    notMySpecies(false),
    gradArray1(npart), gradArray2(npart), 
    temp1(simInfo.getNPart()), temp2(simInfo.getNPart()), nerror(0) {
    //gradArray(npart), mat2(npart,npart),
    //gradMatrix(npart,npart), grad2Matrix(npart,npart) {
  for (unsigned int i=0; i<matrix.size(); ++i)  {
    matrix[i] = new Matrix(npart,npart,ColMajor());
  }
  std::cout << "WireNodes with temperature = "
            << temperature << std::endl;
  double tempp=temperature/(1.0+EPSILON); //Larger beta (plus).
  double tempm=temperature/(1.0-EPSILON); //Smaller beta (minus).
  pg=new PeriodicGaussian(mass*temperature,cell.a[0],
                            (int)(50*cell.a[0]*sqrt(mass*temperature)));
  pgm=new PeriodicGaussian(mass*tempm,cell.a[0],
                             (int)(50*cell.a[0]*sqrt(mass*tempm)));
  pgp=new PeriodicGaussian(mass*tempp,cell.a[0],
                             (int)(50*cell.a[0]*sqrt(mass*tempp)));
  if (useUpdates) {
    updateObj = new MatrixUpdate(maxMovers,maxlevel,npart,matrix,*this);
  }
}

WireNodes::~WireNodes() {
  delete pg; delete pgm; delete pgp;
}

double WireNodes::evaluate(const VArray &r1, const VArray &r2, 
                          const int islice) {
  Matrix& mat(*matrix[islice]);
  mat=0;
  for(int jpart=0; jpart<npart; ++jpart) {
    Vec rj=r1(jpart+ifirst);
    double rj2=0; for (int i=1; i<NDIM; ++i) rj2+=rj[i]*rj[i];
    for(int ipart=0; ipart<npart; ++ipart) {
      // Free particle in idim=0.
      Vec delta(r1(jpart+ifirst)-r2(ipart+ifirst));
      cell.pbc(delta);
      double ear2=(*pg)(fabs(delta[0]));
      //for (int i=0; i<NDIM; ++i) ear2*=(*pg[i])(fabs(delta[i]));
      mat(ipart,jpart)=ear2;
      // SHO in idim>0.
      Vec ri=r2(ipart+ifirst);
      double ri2=0, rirj=0;
      for (int i=1; i<NDIM; ++i) {
        ri2+=ri[i]*ri[i];
        rirj+=ri[i]*rj[i];
      }
      mat(ipart,jpart)*= exp(-c*((ri2+rj2)*coshwt-2.0*rirj));
    }
  }
  // Calculate determinant and inverse.
  int info=0;//LU decomposition
  DGETRF_F77(&npart,&npart,mat.data(),&npart,ipiv.data(),&info);
  if (info!=0) {
    std::cout << "BAD RETURN FROM ZGETRF!!!!" << std::endl;
    nerror++;
    if (nerror>1000) {
      std::cout << "too many errors!!!!" << std::endl;
      std::exit(-1);
    }
  }
  double det = 1;
  for (int i=0; i<npart; ++i) {
    det*= mat(i,i); 
    det *= (i+1==ipiv(i))?1:-1;
  }
  DGETRI_F77(&npart,mat.data(),&npart,ipiv.data(),work.data(),&lwork,&info);
  if (info!=0) {
    std::cout << "BAD RETURN FROM ZGETRI!!!!" << std::endl;
    nerror++;
    if (nerror>1000) {
      std::cout << "too many errors!!!!" << std::endl;
      std::exit(-1);
    }
  }
  return det;
}

void WireNodes::evaluateDotDistance(const VArray &r1, const VArray &r2,
         const int islice, Array &d1, Array &d2) {
  PeriodicGaussian* pgSave;
  pgSave=pg;
  double tauSave=tau;
  // Calculate gradient with finite differences.
  // First calculate distance at smaller beta.
  tau = tauSave*(1.-EPSILON);
  pg=pgm;
  evaluate(r1, r2, islice);
  evaluateDistance(r1, r2, islice, temp1, temp2);
  // Next calculate distance at larger beta.
  tau = tauSave*(1.+EPSILON);
  pg=pgp;
  evaluate(r1, r2, islice);
  evaluateDistance(r1, r2, islice, d1, d2);
  // Now use finite differences.
  double denom=1./(2*tau*EPSILON);
  d1 = denom*(d1-temp1);
  d2 = denom*(d2-temp2);
  // Restore pg and tau to proper temperature.
  pg=pgSave;
  tau=tauSave;
}

void WireNodes::evaluateDistance(const VArray& r1, const VArray& r2,
                              const int islice, Array& d1, Array& d2) {
  d1=200; d2=200; // Initialize distances to a very large value.
  const Matrix& mat(*matrix[islice]);
  for (int jpart=0; jpart<npart; ++jpart) {
    Vec logGrad=0.0;
    Vec rj=r1(jpart+ifirst);
    double rj2=0; for (int i=1; i<NDIM; ++i) rj2+=rj[i]*rj[i];
    Vec bigTerm=0.0;
    for (int ipart=0; ipart<npart; ++ipart) {
      // Free particle in idim=0.
      Vec delta=r1(jpart+ifirst)-r2(ipart+ifirst);
      cell.pbc(delta);
      Vec grad; grad=0.0;
      grad[0]=(*pg).grad(fabs(delta[0]));
      // SHO in idim>0.
      Vec ri=r2(ipart+ifirst);
      double ri2=0, rirj=0;
      for (int i=1; i<NDIM; ++i) {
        ri2+=ri[i]*ri[i];
        rirj+=ri[i]*rj[i];
      }
      double ear2=(*pg)(fabs(delta[0]));
      for (int i=1; i<NDIM; ++i) {
        grad[i]=2*c*(ri[i]-coshwt*rj[i])*ear2;
      }
      grad*=exp(-c*((ri2+rj2)*coshwt-2*rirj));
      if (ipart==jpart) {
        bigTerm=grad/(ear2*exp(-c*((ri2+rj2)*coshwt-2.0*rirj)));
      }
      logGrad+=mat(jpart,ipart)*grad;
    }
    //logGrad-=bigTerm;
    gradArray1(jpart)=logGrad;
    d1(jpart+ifirst)=sqrt(2*mass/((dot(logGrad,logGrad)+1e-15)*tau));
  }
  for (int ipart=0; ipart<npart; ++ipart) {
    Vec logGrad=0.0;
    Vec ri=r2(ipart+ifirst);
    double ri2=0; for (int i=1; i<NDIM; ++i) ri2+=ri[i]*ri[i];
    Vec bigTerm=0.0;
    for(int jpart=0; jpart<npart; ++jpart) {
      // Free particle in idim=0.
      Vec delta=r2(ipart+ifirst)-r1(jpart+ifirst);
      cell.pbc(delta);
      Vec grad; grad=0.0;
      grad[0]=(*pg).grad(fabs(delta[0]));
      // SHO in idim>0.
      Vec rj=r1(jpart+ifirst);
      double rj2=0, rirj=0;
      for (int i=1; i<NDIM; ++i) {
        rj2+=rj[i]*rj[i];
        rirj+=ri[i]*rj[i];
      }
      double ear2=(*pg)(fabs(delta[0]));
      for (int i=1; i<NDIM; ++i) {
        grad[i]=2*c*(rj[i]-coshwt*ri[i])*ear2;
      }
      grad*=exp(-c*((ri2+rj2)*coshwt-2*rirj));
      if (ipart==jpart) {
        bigTerm=grad/(ear2*exp(-c*((ri2+rj2)*coshwt-2.0*rirj)));
      }
      logGrad+=mat(jpart,ipart)*grad;
    }
    //logGrad-=bigTerm;
    gradArray2(ipart)=logGrad;
    d2(ipart+ifirst)=sqrt(2*mass/((dot(logGrad,logGrad)+1e-15)*tau));
  }
}

void WireNodes::evaluateGradLogDist(const VArray &r1, const VArray &r2,
       const int islice, VMatrix &gradd1, VMatrix &gradd2, 
       const Array& dist1, const Array& dist2) {
  gradd1=0.; gradd2=0.; 
/*  const Matrix& mat(*matrix[islice]);
  // ipart is the index of the particle in the gradient.
  // jpart is the index of the particle in the distance.
  for (int ipart=0; ipart<npart; ++ipart) {
    for (int jpart=0; jpart<npart; ++jpart) {
      // First treat d1 terms.
      if (ipart==jpart) {
        Mat d2ii=Mat(0.0);
        for(int kpart=0; kpart<npart; ++kpart) {
          Vec delta=r1(ipart+ifirst)-r2(kpart+ifirst);
          cell.pbc(delta);
          Mat d2iik=Mat(0.0);
          for (int i=0; i<NDIM; ++i) {
            for (int j=0; j<NDIM; ++j) {
              if (i==j) {
                d2iik(i,j)=(*pg[i]).d2(fabs(delta[i]))
                          /(*pg[i])(fabs(delta[i]));
              } else {
                d2iik(i,j)=(*pg[i]).grad(fabs(delta[i]))
                          /(*pg[i])(fabs(delta[i])) 
                          *(*pg[j]).grad(fabs(delta[j]))
                          /(*pg[j])(fabs(delta[j]));
                if (delta[i]*delta[j]<0) d2iik(i,j)=-d2iik(i,j);
              }
            }
          }
          d2ii+=mat(ipart,kpart)*d2iik;
        }
        for (int i=0; i<NDIM; ++i) {
          for (int j=0; j<NDIM; ++j) {
            gradd1(ipart+ifirst,jpart+ifirst)(i)=gradArray1(jpart)(j)*d2ii(j,i);
          }
        }
      } else { // ipart!=jpart
        // Two derivatives act on different columns, so you get a 2x2 det.
        Vec g00(0.0),g01(0.0),g10(0.0),g11(0.0);
        for(int kpart=0; kpart<npart; ++kpart) {
          // First derivative acts on particle ipart.
          Vec delta=r1(ipart+ifirst)-r2(kpart+ifirst);
          cell.pbc(delta);
          Vec grad;
          for (int i=0; i<NDIM; ++i) {
            grad[i]=(*pg[i]).grad(fabs(delta[i]))/(*pg[i])(fabs(delta[i]));
            if (delta[i]<0) grad[i]=-grad[i];
          }
          for (int i=0; i<NDIM; ++i) grad*=(*pg[i])(fabs(delta[i]));
          g00 += mat(ipart,kpart)*grad;
          g10 += mat(jpart,kpart)*grad;
          // Next derivative acts on particle jpart.
          delta=r1(jpart+ifirst)-r2(kpart+ifirst);
          cell.pbc(delta);
          for (int i=0; i<NDIM; ++i) {
            grad[i]=(*pg[i]).grad(fabs(delta[i]))/(*pg[i])(fabs(delta[i]));
            if (delta[i]<0) grad[i]=-grad[i];
          }
          for (int i=0; i<NDIM; ++i) grad*=(*pg[i])(fabs(delta[i]));
          g01 += mat(ipart,kpart)*grad;
          g11 += mat(jpart,kpart)*grad;
        }
        for (int i=0; i<NDIM; ++i) {
          for (int j=0; j<NDIM; ++j) {
            gradd1(ipart+ifirst,jpart+ifirst)(i)=gradArray1(jpart)(j)
                                 *(g00(i)*g11(j)-g10(i)*g01(j));
          }
        }
      }
      // Then treat d2 terms.
      // Derivative on row is a sum of derivatives on each column.
      for(int lpart=0; lpart<npart; ++lpart) {
        if (ipart==lpart) {
          // Only one term is non-zero. NEED TO CHECK THESE DERIVATIVES
          Vec delta=r1(ipart+ifirst)-r2(jpart+ifirst);
          cell.pbc(delta);
          Mat d2ii=Mat(0.0);
          for (int i=0; i<NDIM; ++i) {
            for (int j=0; j<NDIM; ++j) {
              if (i==j) {
                d2ii(i,j)=(*pg[i]).d2(fabs(delta[i]))
                         /(*pg[i])(fabs(delta[i]));
              } else {
                d2ii(i,j)=(*pg[i]).grad(fabs(delta[i]))
                         /(*pg[i])(fabs(delta[i])) 
                         *(*pg[j]).grad(fabs(delta[j]))
                         /(*pg[j])(fabs(delta[j]));
                if (delta[i]*delta[j]>=0) d2ii(i,j)=-d2ii(i,j);
              }
            }
          }
          d2ii*=mat(ipart,jpart);
          for (int i=0; i<NDIM; ++i) {
            for (int j=0; j<NDIM; ++j) {
              gradd2(ipart+ifirst,jpart+ifirst)(i)=gradArray2(jpart)(j)*d2ii(j,i);
            }
          }
        } else { // ipart!=lpart
          // Two derivatives act on different columns, so you get a 2x2 det.
          Vec g00(0.0),g01(0.0),g10(0.0),g11(0.0);
          for(int kpart=0; kpart<npart; ++kpart) {
            // First derivative acts on particle ipart.
            Vec delta=r1(ipart+ifirst)-r2(kpart+ifirst);
            cell.pbc(delta);
            Vec grad;
            for (int i=0; i<NDIM; ++i) {
              grad[i]=(*pg[i]).grad(fabs(delta[i]))/(*pg[i])(fabs(delta[i]));
              if (delta[i]<0) grad[i]=-grad[i];
            }
            for (int i=0; i<NDIM; ++i) grad*=(*pg[i])(fabs(delta[i]));
            g00 += mat(ipart,kpart)*grad;
            g10 += mat(lpart,kpart)*grad;
          }
          // Next derivative acts on particle jpart.
          Vec delta=r2(jpart+ifirst)-r1(lpart+ifirst);
          cell.pbc(delta);
          Vec grad(0.0);
          for (int i=0; i<NDIM; ++i) {
            grad[i]=(*pg[i]).grad(fabs(delta[i]))/(*pg[i])(fabs(delta[i]));
            if (delta[i]<0) grad[i]=-grad[i];
          }
          for (int i=0; i<NDIM; ++i) grad*=(*pg[i])(fabs(delta[i]));
          g01 = mat(ipart,jpart)*grad;
          g11 = mat(lpart,jpart)*grad;
          for (int i=0; i<NDIM; ++i) {
            for (int j=0; j<NDIM; ++j) {
              gradd2(ipart+ifirst,jpart+ifirst)(i)=gradArray2(jpart)(j)
                                   *(g00(i)*g11(j)-g10(i)*g01(j));
            }
          }
        }
      }
      // Put in prefactors.
      gradd1(ipart+ifirst,jpart+ifirst)
        *= -(tau/mass)*dist1(jpart+ifirst)*dist1(jpart+ifirst);
      gradd2(ipart+ifirst,jpart+ifirst)
        *= -(tau/mass)*dist2(jpart+ifirst)*dist2(jpart+ifirst);
      // Then add linear (j independent) contribution.
      gradd1(ipart+ifirst,jpart+ifirst)+=gradArray1(ipart);
      gradd2(ipart+ifirst,jpart+ifirst)+=gradArray1(ipart);
    }
  }*/
}

/*int main(int argc, char** argv) {
  const double SCALE_RAND=1./RAND_MAX;
  const double EPS=1e-4;
  std::cout << "Test free particle nodes" << std::endl;
  const int npart=(int)pow(2,NDIM);
  Species *species=new Species("e",npart,1.0,-1.0,2,true);  
  species->ifirst=0;
  std::vector<Species*> speciesList(1), speciesIndex(npart);
  speciesList[0]=species;
  for (int i=0; i<npart; ++i) speciesIndex[i]=species;
  const double temperature=0.01, tau=0.01;
  const int nslice = (int)(1.0/(temperature*tau));
  SuperCell *cell=new SuperCell(SuperCell::Vec(10.));
  cell->computeRecipricalVectors();
  SimulationInfo simInfo(cell,npart,speciesList,speciesIndex,temperature,
                         tau,nslice);
  WireNodes nodes(simInfo, *species, temperature, 4);
  // Put particles on a grid with random displacements.
  WireNodes::VArray r1(npart), r2(npart);
  for (int ipart=0; ipart<npart; ++ipart) {
    for (int idim=0; idim<NDIM; ++idim) {
      int i=(ipart/(int)pow(2,idim))&1;
      r1(ipart)[idim]=r2(ipart)[idim]=2.5-5.0*i;
      r1(ipart)[idim]+=1.0*(SCALE_RAND*rand()-0.5);
      r2(ipart)[idim]+=1.0*(SCALE_RAND*rand()-0.5);
    }
  }
  WireNodes::Vec delta=r1(0.0)-r2(0.0);
  nodes.evaluate(r1,r2,0);
  double d=nodes.evaluateDistance(r1,r2,0);
  std::cout << "Distance to node " << d << std::endl;
  // Now check forces.
  WireNodes::VArray f(npart);
  nodes.evaluateGradLogDist(r1,r2,0,f,d);
  for (int ipart=0; ipart<npart; ++ipart) {
    for (int idim=0; idim<NDIM; ++idim) {
      std::cout << ipart << "[" << idim << "] = " << f(ipart)[idim] << " ";
      r1(ipart)[idim]+=EPS;
      nodes.evaluate(r1,r2,0);
      double dp=nodes.evaluateDistance(r1,r2,0);
      r1(ipart)[idim]-=2*EPS;
      nodes.evaluate(r1,r2,0);
      double dm=nodes.evaluateDistance(r1,r2,0);
      r1(ipart)[idim]+=EPS;
      std::cout << "(numerical derivative: " 
                << (dp-dm)/(2*EPS*d) << ")" << std::endl;
    }
  }
  return 0;
}*/

const double WireNodes::EPSILON=1e-6;

WireNodes::MatrixUpdate::MatrixUpdate(int maxMovers, int maxlevel, int npart,
    std::vector<Matrix*> &matrix, const WireNodes &wireNodes)
  : wireNodes(wireNodes), maxMovers(maxMovers), npart(npart),
    newMatrix((int)(pow(2,maxlevel)+0.1)+1),
    phi((int)(pow(2,maxlevel)+0.1)+1),
    bvec((int)(pow(2,maxlevel)+0.1)+1),
    smallDet(maxMovers,maxMovers),
    matrix(matrix),
    ipiv(maxMovers),lwork(maxMovers*maxMovers),work(lwork),
    isNewMatrixUpdated(false) {
  for (unsigned int i=0; i<matrix.size(); ++i)  {
    newMatrix[i] = new Matrix(npart,npart,ColMajor());
    phi[i] = new Matrix(npart,maxMovers,ColMajor());
    bvec[i] = new Matrix(npart,maxMovers,ColMajor());
  }
}

double WireNodes::MatrixUpdate::evaluateChange(
    const DoubleMLSampler &sampler, int islice) {
  isNewMatrixUpdated=false;
  // Get info on moving paths.
  const Beads<NDIM>& sectionBeads2=sampler.getSectionBeads(2);
  const Beads<NDIM>& movingBeads1=sampler.getMovingBeads(1);
  index1=&sampler.getMovingIndex(1); 
  nMoving=index1->size();
  // Compute new slater matrix elements for moving particles.
  for (int jmoving=0; jmoving<nMoving; ++jmoving) {
    //Vec rj=r1(jpart+ifirst);
    Vec rj=movingBeads1(jmoving,islice);
    double rj2=0; for (int i=1; i<NDIM; ++i) rj2+=rj[i]*rj[i];
    for (int ipart=0; ipart<npart; ++ipart) {
      Vec ri=sectionBeads2(ipart+wireNodes.ifirst,islice);
      // Free particle in idim=0.
      Vec delta(rj-ri);
      wireNodes.cell.pbc(delta);
      double ear2=(*wireNodes.pg)(fabs(delta[0]));
      (*phi[islice])(ipart,jmoving)=ear2;
      // SHO in idim>0.
      double ri2=0, rirj=0;
      for (int i=1; i<NDIM; ++i) {
        ri2+=ri[i]*ri[i];
        rirj+=ri[i]*rj[i];
      }
      (*phi[islice])(ipart,jmoving)*= exp(-wireNodes.c*((ri2+rj2)
                               *wireNodes.coshwt-2.0*rirj));
    }
    for (int imoving=0; imoving<nMoving; ++imoving) {
      int ipart=(*index1)(imoving)-wireNodes.ifirst;
      (*bvec[islice])(ipart,jmoving)=0;
      for (int k=0; k<npart; ++k) {
        (*bvec[islice])(ipart,jmoving)
          += (*phi[islice])(k,jmoving)*(*matrix[islice])(ipart,k);
      }
    }
  }
  // Compute change in the slater determinant.
  for (int jmoving=0; jmoving<nMoving; ++jmoving) {
    for (int imoving=0; imoving<nMoving; ++imoving) {
      int ipart=(*index1)(imoving)-wireNodes.ifirst;
      smallDet(imoving,jmoving)=(*bvec[islice])(ipart,jmoving);
    }
  }
  // Calculate determinant and inverse.
  int info=0;//LU decomposition
  DGETRF_F77(&nMoving,&nMoving,smallDet.data(),&maxMovers,ipiv.data(),&info);
  if (info!=0) std::cout << "BAD RETURN FROM ZGETRF!!!!" << std::endl;
  double det = 1;
  for (int i=0; i<nMoving; ++i) {
    det *= smallDet(i,i); 
    det *= (i+1==ipiv(i))?1:-1;
  }
  return det;
}

void WireNodes::MatrixUpdate::evaluateNewInverse(const int islice) {
  //isNewMatrixUpdated=true;
}

void WireNodes::MatrixUpdate::evaluateNewDistance(const int islice, 
    Array &d1, Array &d2) {
}

void WireNodes::MatrixUpdate::acceptLastMove(int nslice) {
  //if (isNewMatrixCurrent) ... else
  for (int islice=1; islice<nslice-1; ++islice) {
    for (int jmoving=0; jmoving<nMoving; ++jmoving) {
      int jpart=(*index1)(jmoving)-wireNodes.ifirst;
      double bjjinv=0;
      for (int k=0; k<npart; ++k) {
        bjjinv += (*phi[islice])(k,jmoving)*(*matrix[islice])(jpart,k);
      }
      bjjinv=1./bjjinv;
      for (int k=0; k<npart; ++k) {
        (*matrix[islice])(jpart,k) *= bjjinv;
      } 
      for (int ipart=0; ipart<npart; ++ipart) {
        if (ipart==jpart) continue;
        double bij=0;
        for (int k=0; k<npart; ++k) {
          bij += (*phi[islice])(k,jmoving)*(*matrix[islice])(ipart,k);
        }
        for (int k=0; k<npart; ++k) {
          (*matrix[islice])(ipart,k) -= (*matrix[islice])(jpart,k)*bij;
        }
      }
    }
  }
}
